
/**
 * Title:        Order Processing Server Example<p>
 * Description:  This is an example of an email based XML order processing
 * server that was written for the book Strategic XML, published by
 * Sams publishing.<p>
 * Copyright:    Copyright (c) W. Scott Means<p>
 * Company:      Strategic XML<p>
 * @author W. Scott Means
 * @version 1.0
 */
package com.strategicxml.examples.OrderServer;

import java.util.*;
import java.io.*;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;

import com.jthomas.pop.*;

public class OrderServer implements Runnable {
  //
  // static constants
  //

  /**
   * Class name of the default SAX parser to use.
   */
  private static final String DEFAULT_SAX_DRIVER_CLASS =
      "org.apache.xerces.parsers.SAXParser";

  /**
   * Default server configuration file URL.
   */
  private static final String DEFAULT_CONFIG_FILE_URL =
      "OrderServerConfig.xml";

  /**
   * Order database XML URL.
   */
  private static final String DEFAULT_ORDER_DB_URL =
      "Orders.xml";

  /**
   * Default number of milliseconds to wait before checking POP box again.
   */
  private static final long DEFAULT_CHECK_INTERVAL = 30000;

  public static void main(String[] args) {
    OrderServer osRun = new OrderServer();

    if (osRun.runOrderService(args)) {
      // continue running the service until the user hits a key
      try {
        System.in.read();
      } catch (IOException ioe) {
        System.err.println(ioe);
      }

      osRun.stopOrderService();
    }
  }

  //
  // instance members
  //
  protected Thread m_thrdService;
  protected boolean m_fStopService = false;
  protected Document m_docServerConfig;
  protected SAXBuilder m_sb;
  protected String m_strOrderDBFName = DEFAULT_ORDER_DB_URL;
  protected File m_fileOrderDB;
  protected long m_lCheckInterval = DEFAULT_CHECK_INTERVAL;
  protected String m_strPOPServer;
  protected String m_strPOPUser;
  protected String m_strPOPPass;

  /**
   * OrderServer
   *
   * Empty constructor for SMTP order server service.
   */
  public OrderServer() {
    m_sb = new SAXBuilder(DEFAULT_SAX_DRIVER_CLASS);
  }

  /**
   * Intializes and starts executing the order service daemon. Uses Java threads
   * to start a background thread that monitors the POP3 box for incoming
   * messages.
   */
  public synchronized boolean runOrderService(String[] args)
  {
    if (m_thrdService != null) {
      // service thread is already running
      return false;
    }

    try {
      // first, open the service configuration file
      if (args.length > 0) {
        m_docServerConfig = m_sb.build(args[0]);
      } else {
        m_docServerConfig = m_sb.build(DEFAULT_CONFIG_FILE_URL);
      }
    } catch (JDOMException jde) {
      System.err.println(jde);

      return false;
    }

    if (m_docServerConfig == null) {
      return false;
    }

    // initialize from configuration file
    m_strPOPServer = m_docServerConfig.getRootElement().getChild("pop_server").getTextTrim();
    m_strPOPUser = m_docServerConfig.getRootElement().getChild("pop_user").getTextTrim();
    m_strPOPPass = m_docServerConfig.getRootElement().getChild("pop_pass").getTextTrim();
    m_strOrderDBFName = m_docServerConfig.getRootElement().getChild("order_db_file").getTextTrim();
    m_fileOrderDB = new File(m_strOrderDBFName);
    m_fStopService = false;
    m_thrdService = new Thread(this);
    m_thrdService.start();

    return true;
  }

  /**
   * This method will stop the order service thread.
   */
  public synchronized void stopOrderService()
  {
    m_fStopService = true;
  }

  /**
   * Entry point of order server daemon thread.
   */
  public void run()
  {
    do {
      System.out.println("Checking server " + m_strPOPServer + "...");

      if (!checkPOPServer()) {
        break;
      }

      System.out.println("Waiting...");
      try {
        synchronized (this) {
          wait(m_lCheckInterval);
        }
      } catch (InterruptedException ie) {
        m_fStopService = true;
      }
      System.out.println("done.");
    } while (!m_fStopService);

    synchronized(this) {
      m_thrdService = null;
    }
  }

  public boolean checkPOPServer()
  {
    pop3 pop = new pop3(m_strPOPServer, m_strPOPUser, m_strPOPPass);

    popStatus status = pop.connect();

    if (status.OK()) {
      status = pop.login();

      if (status.OK()) {
        int cMsgs = pop.get_TotalMsgs();

        if (status.OK()) {
          for (int i = 1; i <= cMsgs; i++) {
            popStatus statMsg = pop.retr(i);

            if (statMsg.OK()) {
              String [] astrMsg = statMsg.Responses();

              processMessage(astrMsg);
            }
          }

          for (int i = 1; i <= cMsgs; i++) {
            pop.dele(i);
          }

          status = pop.quit();
        }
      }
    }

    return status.OK();
  }

  /**
   * This method parses the incoming message and if all goes well, adds it to
   * the order database document.
   */
  public void processMessage(String [] astrMsg)
  {
    // first, capture the email message headers in a hash map
    HashMap hmHeaders = new HashMap();

    int i;

    for (i = 0; i < astrMsg.length && astrMsg[i].length() > 0; i++) {
      int iColon = astrMsg[i].indexOf(':');

      if (iColon > 0) {
        hmHeaders.put(astrMsg[i].substring(0, iColon).trim().toLowerCase(),
            astrMsg[i].substring(iColon+1).trim());
      }
    }

    // print some status info.
    System.out.println("processing message from " + hmHeaders.get("from"));

    // now, build the XML document
    StringBuffer sb = new StringBuffer();

    for (i++; i < astrMsg.length; i++) {
      sb.append(astrMsg[i] + "\n");
    }

    try {
      Document docOrder = m_sb.build(new StringReader(sb.toString()));

      if (docOrder != null) {
        // OK, let's save some of the headers from the order email
        Element elEmail = new Element("email_info");
        Element el = new Element("from");

        el.setText((String)hmHeaders.get("from"));

        elEmail.addContent(el);

        el = new Element("date");
        el.setText((String)hmHeaders.get("date"));

        elEmail.addContent(el);

        docOrder.getRootElement().addContent(elEmail);

        Document docDB;

        try {
          docDB = m_sb.build(new FileInputStream(m_fileOrderDB));
        } catch (Exception e) {
          // either the order database document doesn't exist, or
          // there was an error while parsing it
          docDB = null;
        }

        if (docDB == null) {
          docDB = new Document(new Element("order_database"));

          docDB.getRootElement().addContent(new Element("orders"));
        }

        Element elOrders = docDB.getRootElement().getChild("orders");

        Element elRoot = docOrder.getRootElement();
        docOrder.setRootElement(new Element("temp"));

        elOrders.addContent(elRoot);

        try {
          XMLOutputter xo = new XMLOutputter();

          FileOutputStream fos = new FileOutputStream(m_fileOrderDB);
          xo.output(docDB, fos);
          fos.close();
        } catch (IOException ioe) {
          System.err.println(ioe);
        }
      }
    } catch (JDOMException jde) {
      System.err.println(jde);
    }
  }
}